from PyTASTileTypeLight import *
import numpy as np
import random
import copy
import xml.etree.ElementTree as ET
import xml.dom.minidom
from scipy.spatial.transform import Rotation
import svgwrite
from svgwrite import cm, mm, px
from svgwrite import pt as svgPt
from reportlab.pdfgen import canvas

DISPLAY_ALL = 0
DISPLAY_GLUES = 1
DISPLAY_NO_GLUES = 2

###################################################
# Currently hard-coded parameters for drawing

# Glue 'functions'
in1 = 'in1'
in2 = 'in2'
out1 = 'out1'
out2 = 'out2'
out3 = 'out3'
out4 = 'out4'
out5 = 'out5'
intau = 'intau'
outtau = 'outtau'
outtau1 = 'outtau1'
outtau2 = 'outtau2'
linktau = 'linktau'  # linktau is not currently fully supported

glue_functions = [in1, in2, out1, out2, out3, out4, out5, intau, outtau, outtau1, outtau2, linktau]

glue_markings = {'in1': {N: 'v', E: '<', S: '^', W: '>'},
                 'in2': {N: 'v', E: '<', S: '^', W: '>'},
                 'out1': {N: '^', E: '>', S: 'v', W: '<'},
                 'out2': {N: '^', E: '>', S: 'v', W: '<'},
                 'out3': {N: '^', E: '>', S: 'v', W: '<'},
                 'out4': {N: '^', E: '>', S: 'v', W: '<'},
                 'out5': {N: '^', E: '>', S: 'v', W: '<'},
                 'intau': {N: 'v', E: '<', S: '^', W: '>'},
                 'outtau': {N: '^', E: '>', S: 'v', W: '<'},
                 'outtau1': {N: '^', E: '>', S: 'v', W: '<'},
                 'outtau2': {N: '^', E: '>', S: 'v', W: '<'},
                 'linktau': {N: '^', E: '>', S: '^', W: '>'}}

#squareDim = 10
strokeWidth = 1
svgUnits = px
skinnyFactor = 0.7
textStyle = "font-family:Arial"
defaultFillColor = 'grey'
pdfFont = "Courier"#"Helvetica"
pdfFontBold = "Helvetica-Bold"
#partial_example_color_list = ["black", "black", "green", "red", "blue", "yellow", "pink", "aqua", "lime", "olive", "black"]

# The following is a dictionary of default glue colors which can be extended and/or modified
slat_color_mapping = {'vertical': 'grey', 'horizontal': 'grey', in1: 'lime', in2: 'yellow',
                      out1: 'pink', out2: 'aqua', out3: 'blue', out4: 'blue', out5: 'blue', outtau: 'red', outtau1: 'red', outtau2: 'red',
                      intau: 'red', linktau: 'olive',
                      'crossing_glue_color': 'red', 'linking_glue_color': 'olive'}




outline_pen = None

DIR_OFFSETS = {
    'N':  np.array([ 0,  1,  0]),
    'S':  np.array([ 0, -1,  0]),
    'E':  np.array([ 1,  0,  0]),
    'W':  np.array([-1,  0,  0]),
    'U':  np.array([ 0,  0,  1]),
    'D':  np.array([ 0,  0, -1])
}

REV_DIR_OFFSETS = {
    ( 0,  1,  0):   'N',
    ( 0, -1,  0):   'S',
    ( 1,  0,  0):   'E',
    (-1,  0,  0):   'W',
    ( 0,  0,  1):   'U',
    ( 0,  0, -1):   'D',
}

DIR_OPPS = {
    'N': 'S',
    'S': 'N',
    'E': 'W',
    'W': 'E',
    'U': 'D',
    'D': 'U',
}

glue_offset = {'N':  (1,0),
               'E':  (2,1),
               'S':  (1,2),
               'W':  (0,1),
               'U':  (1,1),
               'D':  (1,1),
               'NE': (2,0),
               'SE': (2,2),
               'SW': (0,2),
               'NW': (0,0),
}

def GetFontSizeToFitTextInBounds(text, font, dc, width, height):
    retFont = font
    (w, pixelSize) = retFont.GetPixelSize()
    bContinue = True
    while bContinue == True:
        dc.SetFont(retFont)
        (w, h) = dc.GetTextExtent(text)
        if ((w <= width) and (h <= height)) or pixelSize <= 2:
            bContinue = False
        else:
            pixelSize = pixelSize - 2
            retFont.SetPixelSize((0, pixelSize))
    #(w, pixelSize) = retFont.GetPixelSize()
    #print("pixelSize: " + str(pixelSize))
    return retFont

def make_3D(coords):
    ret_coords = coords
    if len(ret_coords) < 3:
        ret_coords = (coords[0], coords[1], 0)
    return ret_coords

def get_label_complement(label):
    if label[-1] == '*':
        return label[0:-1]
    else:
        return label + '*'

def labels_complement(l1, l2):
    return l1 == l2 + '*' or l2 == l1 + '*'

class SlatType:
    def __init__(self, name, color=None, blocks=None, colors=None, domains=None, xml_node=None, dim_restrictions=(None, None, None)):
        self.name = name
        self.color = color
        self.base_type = None # Rotated versions will maintain a pointer to the 'base' (unrotated) type
        self.concentration = 1.0
        self.reference_point = None
        self.bounding_box = None
        self.b_seed_only = False
        self.placement_hint = None
        self.dim_restrictions = dim_restrictions
        if xml_node is None:
            self.blocks = blocks or [(0, 0, 0)]
            self.colors = colors or {}
            self.domains = domains or {}
        else:
            self.fromXML(xml_node)
        self.init_color()
        self.ensure_3D()
        self.ComputeReferencePoint()
        self.ComputePerimeterOffsets(self.dim_restrictions)

    def __str__(self):
        return f'Slat type {self.name}: blocks: {self.blocks}, domains: {self.domains}'
    
    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        if self.name != other.name:
            return False
        if self.domains.keys() != other.domains.keys():
            return False
        for loc in self.domains:
            if set(self.domains[loc]) != set(other.domains[loc]):
                return False
        if self.blocks != other.blocks:
            return False
        return True

    def __hash__(self):
        n = hash(self.name)
        b = hash(tuple(sorted(list(self.blocks))))
        d = hash(tuple(sorted([(l, tuple(sorted(d))) for l, d in self.domains.items()])))
        return hash((n, b, d))

    def IsHorizontal(self):
        if self.blocks and len(self.blocks) > 1:
            if self.blocks[0][0] != self.blocks[1][0]:
                return True
        return False

    def get_domain_label_list(self):
        domain_label_list = []
        for block in self.blocks:
            if not block in self.domains:
                domain_label_list.append('')
            else:
                domain = self.domains[block][0]
                domain_label_list.append(domain[0])
        return domain_label_list

    def get_location_list(self):
        location_list = []
        if self.blocks:
            for block in self.blocks:
                location_list.append(block)
        return location_list

    def ensure_3D(self):
        new_blocks = []
        for block in self.blocks:
            new_blocks.append(make_3D(block))
        self.blocks = new_blocks
        new_colors = dict()
        for color_coords in self.colors:
            new_colors[make_3D(color_coords)] = self.colors[color_coords]
        self.colors = new_colors
        new_domains = dict()
        for domain_coords in self.domains:
            new_domains[make_3D(domain_coords)] = self.domains[domain_coords]
        self.domains = new_domains

    def init_color(self):
        pass
##        if self.color is None:
##            self.color = "LIGHT GREY"
##        else:
##            wx_color = wx.TheColourDatabase.Find(self.color)
##            if wx_color == False:
##                msg = f"Invalid color '{self.color}' specified for slat type {self.name}"
##                SimUtils.globalApp.OutputMessage(msg)
##                self.color = "LIGHT GREY"
##        color = wx.TheColourDatabase.Find(self.color)
##        self.colorVector = (color.Red(), color.Green(), color.Blue())

    def Copy(self):
        pct = SlatType(name=self.name, color=self.color, dim_restrictions=self.dim_restrictions)
        pct.blocks = copy.deepcopy(self.blocks)
        pct.colors = copy.deepcopy(self.colors)
        pct.domains = copy.deepcopy(self.domains)
        pct.concentration = self.concentration
        pct.reference_point = self.reference_point
        pct.placement_hint = self.placement_hint
        pct.b_seed_only = self.b_seed_only
        pct.ComputePerimeterOffsets(self.dim_restrictions)
        return pct
        
    def ComputeReferencePoint(self):
        min_z = 9999999
        min_y = 9999999
        min_x = 9999999
        
        for block in self.blocks:
            min_z = min(min_z, block[2])
        for block in self.blocks:
            if block[2] == min_z:
                min_y = min(min_y, block[1])
        for block in self.blocks:
            if (block[1] == min_y) and (block[2] == min_z):
                min_x = min(min_x, block[0])
        self.reference_point = (min_x, min_y, min_z)

    def translate(self, vector):
        if len(vector) < 3:
            vector = (vector[0], vector[1], 0)
        new_blocks = []
        self.translation = (self.reference_point[0]+vector[0], self.reference_point[1]+vector[1], self.reference_point[2]+vector[2])
        for block in self.blocks:
            x,y,z = block[0] + vector[0], block[1] + vector[1], block[2] + vector[2]
            new_blocks.append((x,y,z))
        self.blocks = new_blocks

        translated_colors = dict()
        for color in self.colors:
            loc = (color[0] + vector[0], color[1] + vector[1], color[2] + vector[2])
            translated_colors[loc] = self.colors[color]
        self.colors = translated_colors

        translated_domains = dict()
        for domain in self.domains:
            loc = (domain[0] + vector[0], domain[1] + vector[1], domain[2] + vector[2])
            translated_domains[loc] = self.domains[domain]
        self.domains = translated_domains
        self.bounding_box = None

    def translate_reference_point_to(self, coords):
        diff = ( (coords[0] - self.reference_point[0]),
                 (coords[1] - self.reference_point[1]),
                 (coords[2] - self.reference_point[2]) )
        self.translate(diff)

    def Initialize(self):
        b_horizontal = self.IsHorizontal()

        self.ComputeReferencePoint()

        # Translate to (0,0)
        z = 1 if b_horizontal else 0
        self.translate_reference_point_to((0, 0, z))

        # Ensure all domains on horizontal slats end with '*' and none on vertical slats do
        for block in self.domains:
            domain_list = self.domains[block]
            for d in range(len(domain_list)):
                if b_horizontal:                # Make all glues starred
                    if domain_list[d][0][-1] != '*':
                        domain_list[d][0] = f'{domain_list[d][0]}*'
                else:                            # Make all glues unstarred
                    if domain_list[d][0][-1] == '*':
                        domain_list[d][0] = domain_list[d][0][0:-1]

            
    def GetRotatedVersion(self, ccw_rotation_angle):
        dir_arr = ['N','W','S','E']
        rot_amt = int(ccw_rotation_angle / 90)

        pctR = SlatType(self.name + f"(r{str(ccw_rotation_angle)})", self.color)
        blocksR = []
        domainsR = dict()
        for block in self.blocks:
            rot = Rotation.from_euler('z', ccw_rotation_angle, degrees=True)
            rot_coords = rot.apply([block[0], block[1], block[2]])
            blockR = (round(rot_coords[0]), round(rot_coords[1]), round(block[2]))
            blocksR.append(blockR)

            if block in self.domains:
                domain_list = self.domains[block]
                new_domain_list = []
                for domain in domain_list:
                    direction = domain[1]
                    strength = domain[2]
                    if direction in dir_arr:
                        new_domain_list.append((domain[0], dir_arr[(dir_arr.index(direction)+rot_amt)%4], strength))
                    else:
                        new_domain_list.append((domain[0], direction, strength))
                domainsR[blockR] = new_domain_list

        pctR.blocks = blocksR
        pctR.domains = domainsR
        pctR.ComputeReferencePoint()
        pctR.ComputePerimeterOffsets(self.dim_restrictions)
        return pctR

    """
        Compute the list offsets for perimeter locations.
        
        Parameters:
            plane_restrictions : a 3-tuple containing pairs of min and max coords (min is included, max is exluded)
                                 allowed in each dimension. If None, no restriction in that dimension.
    """
    def ComputePerimeterOffsets(self, plane_restrictions=(None, None, None)):
        self.perimeter_offsets = []
        for block in self.blocks:
            for add in [(-1,0,0), (1,0,0), (0,-1,0), (0,1,0), (0,0,-1), (0,0,1)]:
                nbr = (block[0]+add[0], block[1]+add[1], block[2]+add[2])
                if not nbr in self.blocks:
                    if (((plane_restrictions[0] is None) or ((nbr[0] >= plane_restrictions[0][0]) and (nbr[0] < plane_restrictions[0][1]))) and
                        ((plane_restrictions[1] is None) or ((nbr[1] >= plane_restrictions[1][0]) and (nbr[1] < plane_restrictions[1][1]))) and
                        ((plane_restrictions[2] is None) or ((nbr[2] >= plane_restrictions[2][0]) and (nbr[2] < plane_restrictions[2][1])))):
                        self.perimeter_offsets.append(nbr)
       
    def GetMaxBinding(self):
        total_str = 0
        for block in self.domains:
            domain_list = self.domains[block]
            for domain in domain_list:
                total_str = total_str + domain[2]
        return total_str

    def Draw(self, dc, zero_coords, side_len):
        for block in self.blocks:
            dc.SetBrush(wx.Brush(wx.TheColourDatabase.Find(self.color)))
            self.DrawBlock(dc, zero_coords, side_len, block, self.domains)
            
    def DrawBlock(self, dc, zero_coords, side_len, block, domains, colors):
        x_start = zero_coords[0] + (side_len * block[0])
        y_start = zero_coords[1] - (side_len * block[1])

        top_left = wx.Point(x_start, y_start)
        bottom_right = wx.Point(x_start + side_len, y_start + side_len)

        fill_color = "LIGHT GREY"
        if block in colors:
            fill_color = colors[block]
        elif self.color:
            fill_color = self.color
        dc.SetBrush(wx.Brush(wx.TheColourDatabase.Find(fill_color)))
        ##########################dc.SetPen(wx.Brush(wx.TheColourDatabase.Find('red')))
        dc.DrawRectangle(wx.Rect(top_left, bottom_right))

        if block in domains:
            dc.SetBrush(wx.Brush(wx.TheColourDatabase.Find("RED")))
            third = int(side_len/3)
            domain_list = domains[block]
            for domain in domain_list:
                label = domain[0]
                direction = domain[1]
                top_left = wx.Point(x_start+(glue_offset[direction][0]*third), y_start+(glue_offset[direction][1]*third))
                bottom_right = wx.Point(top_left[0]+third, top_left[1]+third)
                if direction == 'U':
                    radius = (bottom_right.x-top_left.x)/2
                    dc.DrawCircle(top_left.x + radius, top_left.y + radius, radius)
                else:
                    dc.DrawRectangle(wx.Rect(top_left, bottom_right))
                dc.DrawLabel(label, wx.Rect(top_left, bottom_right), alignment=wx.ALIGN_CENTER_HORIZONTAL|wx.ALIGN_CENTER_VERTICAL)

    def MakeImage(self, backgroundColor=None, mask=DISPLAY_ALL, pdfCanvas=None, svgCanvas=None, squareDim=None,
                  pageHeight=None, blocks=None, domains=None, translation=(0,0), colors=None):
        if not svgCanvas is None:
            svgGroup = svgCanvas.add(svgCanvas.g())

        color_mapping = slat_color_mapping

        if not blocks:
            blocks = self.blocks
        if not domains:
            domains = self.domains
        if not colors:
            colors = self.colors


        b_horizontal = True if (blocks[0][0] != blocks[1][0]) else False

        for block in blocks:
            (curr_x, curr_y) = (block[0], block[1])

            if b_horizontal:
                x_dim = squareDim
                y_dim = squareDim * skinnyFactor
                x_offset = 0
                y_offset = (squareDim - (squareDim * skinnyFactor))/2
                if block in colors:
                    fillColor = colors[block]
                elif self.color is not None:
                    fillColor = self.color
                elif 'horizontal' in color_mapping:
                    fillColor = color_mapping['horizontal']
                else:
                    fillColor = defaultFillColor
            else:
                x_dim = squareDim * skinnyFactor
                y_dim = squareDim
                x_offset = (squareDim - (squareDim * skinnyFactor))/2
                y_offset = 0
                if block in colors:
                    fillColor = colors[block]
                elif self.color is not None:
                    fillColor = self.color
                elif 'vertical' in color_mapping:
                    fillColor = color_mapping['vertical']
                else:
                    fillColor = defaultFillColor

            domain = None
            if block in domains:
                domain_list = domains[block]
                if len(domain_list) > 0:
                    domain = domain_list[0]
            
##                if not io_loc[2] is None:
##                    fillColor = io_loc[2]
##                elif io_loc[1] in color_mapping:
##                    fillColor = color_mapping[io_loc[1]]

            if not svgCanvas is None:
                x_loc = ((curr_x + translation[0]) * squareDim) + x_offset
                y_loc = pageHeight - ((curr_y+translation[1]+1) * squareDim) + y_offset
                stroke_color = 'white' if fillColor == 'black' else 'black'

                svgSquareGroup = svgGroup.add(svgCanvas.g())
                
                svgSquareGroup.add(svgCanvas.rect(insert=(x_loc*svgUnits, y_loc*svgUnits),
                                      size=(x_dim*svgUnits, y_dim*svgUnits),
                                      fill=fillColor,
                                      stroke=stroke_color,
                                      stroke_width=strokeWidth))

                if (not domain is None) and (not mask == DISPLAY_NO_GLUES):
                    text = domain[0]

                    #print(f'glue label: {text}')
                    textLen = len(text)
                    if textLen < 3:
                        #text = f' {text} '
                        textSize = x_dim/textLen*.7
                        textLength = textLen*textSize*0.7
                    else:
                        textSize = x_dim/textLen
                        textLength = x_dim*0.7
                    xOffset = x_loc + (x_dim/2) - textLength/2
                    text = svgCanvas.text(text,insert=(xOffset*svgUnits, (y_loc+y_dim*.6)*svgUnits),
                                          textLength=textLength*svgUnits, font_size=textSize*svgUnits,
                                          fill=stroke_color, style=textStyle)
                    svgSquareGroup.add(text)

            if not pdfCanvas is None:
                pdfCanvas.setFillColor(fillColor)
                x_loc = (curr_x + translation[0]) * squareDim + x_offset
                y_loc = (curr_y + translation[1]) * squareDim + y_offset
                pdfCanvas.rect(x_loc, y_loc, x_dim, y_dim, stroke=strokeWidth, fill=1)

                if (not domain is None) and (not mask == DISPLAY_NO_GLUES):
                    text = domain[0]
                    if len(text) < 3:
                        text = f' {text} '
                    textLen = len(text)
                    fontSize = x_dim / textLen

                    y_loc = y_loc + y_dim/2.5
                    x_loc = x_loc + x_dim/4
                    
                    textobject = pdfCanvas.beginText(x_loc, y_loc)
                    textobject.setFont(pdfFont, fontSize)
                    if fillColor == 'black':
                        textobject.setFillColorRGB(255,255,255)
                    else:
                        textobject.setFillColorRGB(0,0,0)
                    textobject.textOut(text)
                    pdfCanvas.drawText(textobject)

    def toXML(self, xml_parent=None, color_mapping={}):
        #print(f'colors: {self.colors}')
        if xml_parent is None:
            xml_parent = ET.Element('PolyominoType')
            elt_ptype = xml_parent
        else:
            elt_ptype = ET.SubElement(xml_parent, 'PolyominoType')

        if not self.name is None:
            elt_name = ET.SubElement(elt_ptype, 'name')
            elt_name.text = self.name
        if not self.color is None:
            elt_color = ET.SubElement(elt_ptype, 'color')
            elt_color.text = color_mapping[self.color] if self.color in color_mapping else self.color
        if not self.concentration is None:
            elt_concentration = ET.SubElement(elt_ptype, 'concentration')
            elt_concentration.text = str(self.concentration)
        if not self.placement_hint is None:
            elt_placement_hint = ET.SubElement(elt_ptype, 'placement_hint')
            elt_placement_hint.text = str(self.placement_hint)
        if self.b_seed_only == True:
            elt_seed_only = ET.SubElement(elt_ptype, 'b_seed_only')
            elt_seed_only.text = str('True')
            
        elt_blocks = ET.SubElement(elt_ptype, 'blocks')
        for block in self.blocks:
            elt_block = ET.SubElement(elt_blocks, 'block')
            elt_coords = ET.SubElement(elt_block, 'coords')
            elt_coords.text = str(block)
            if block in self.colors:
                elt_color = ET.SubElement(elt_block, 'color')
                elt_color.text = color_mapping[self.colors[block]] if self.colors[block] in color_mapping else self.colors[block]
            if block in self.domains:
                elt_domains = ET.SubElement(elt_block, 'domains')
                domain_list = self.domains[block]
                for domain in domain_list:
                    elt_domain = ET.SubElement(elt_domains, 'domain')
                    elt_label = ET.SubElement(elt_domain, 'label')
                    elt_label.text = domain[0]
                    elt_direction = ET.SubElement(elt_domain, 'direction')
                    elt_direction.text = domain[1]
                    if (len(domain) > 2) and (not domain[2] == 1):
                        elt_strength = ET.SubElement(elt_domain, 'strength')
                        elt_strength.text = str(domain[2])
        return xml_parent
    
    def fromXML(self, xml_parent):
        if (not xml_parent.tag == 'SlatType') and (not xml_parent.tag == 'PolyominoType'):
            print("Invalid XML node for SlatType")
        else:
            blocks = []
            self.colors = dict()
            self.domains = dict()
            for elt_child in xml_parent:
                if elt_child.tag == 'name':
                    self.name = elt_child.text
                elif elt_child.tag == 'color':
                    if not elt_child.text is None:
                        self.color = elt_child.text.lower()
                elif elt_child.tag == 'concentration':
                    self.concentration = float(elt_child.text)
                elif elt_child.tag == 'placement_hint':
                    self.placement_hint = eval(elt_child.text)
                elif elt_child.tag == 'b_seed_only':
                    self.b_seed_only = True if elt_child.text == 'True' else False
                elif elt_child.tag == 'blocks':
                    # get blocks
                    for elt_block in elt_child:
                        if elt_block.tag == 'block':
                            for elt_child2 in elt_block:
                                if elt_child2.tag == 'coords':
                                    coords = eval(elt_child2.text)
                                    blocks.append(coords)
                                elif elt_child2.tag == 'color':
                                    self.colors[coords] = elt_child2.text
                                elif elt_child2.tag == 'domains':
                                    # get domains
                                    self.domains[coords] = []
                                    for elt_domain in elt_child2:
                                        if elt_domain.tag == 'domain':
                                            label = None
                                            direction = None
                                            strength = 1
                                            for elt_child3 in elt_domain:
                                                if elt_child3.tag == 'label':
                                                    label = elt_child3.text
                                                elif elt_child3.tag == 'direction':
                                                    direction = elt_child3.text
                                                elif elt_child3.tag == 'strength':
                                                    strength = int(elt_child3.text)
                                            if (not label is None) and (not coords is None):
                                                self.domains[coords].append((label, direction, strength))
            self.blocks = blocks

    def getBoundingBox(self, blocks=None):
        if self.bounding_box is None:
            min_x, min_y, min_z =  999999, 999999,  999999
            max_x, max_y, max_z = -999999,-999999, -999999
            bEmpty = True
            
            if blocks is None:
                blocks = self.blocks
    
            for block in blocks:
                bEmpty = False
                if block[0] < min_x:
                    min_x = block[0]
                if block[0] > max_x:
                    max_x = block[0]
                if block[1] < min_y:
                    min_y = block[1]
                if block[1] > max_y:
                    max_y = block[1]
                if block[2] < min_z:
                    min_z = block[2]
                if block[2] > max_z:
                    max_z = block[2]
    
            if bEmpty == True:
                self.bounding_box = ((0,0), (0,0), (0,0))
            else:
                self.bounding_box = ((min_x, min_y, min_z), (max_x, max_y, max_z))
        return self.bounding_box

    def Get2DOutline(self, blocks=None):
        if blocks == None:
            blocks = self.blocks
        offsets = {(1,0) :((1,-1),(1,0)),
                   (-1,0):((0,-1),(0,0)),
                   (0,1) :((0,0),(1,0)),
                   (0,-1):((0,-1),(1,-1))}
        outer_edges = []
        for block in blocks:
            for offset in offsets:
                nbr_block = (block[0]+offset[0], block[1]+offset[1], block[2])
                if not nbr_block in blocks:
                    (start_pt, end_pt) = offsets[offset]
                    start_pt = (start_pt[0]+block[0], start_pt[1]+block[1])
                    end_pt = (end_pt[0]+block[0], end_pt[1]+block[1])
                    outer_edges.append(((start_pt[0], start_pt[1]), (end_pt[0], end_pt[1])))
        return outer_edges
            
class Slat:
    def __init__(self, slatType, xml_node=None, slatTypes=None):
        if xml_node is None:
            self.slatType = slatType
            self.color = self.slatType.color
            self.blocks = copy.deepcopy(self.slatType.blocks)
            self.colors = copy.deepcopy(self.slatType.colors)
            self.domains = copy.deepcopy(self.slatType.domains)
            self.translation = (0,0,0)
        else:
            self.translation = (0,0,0)
            self.fromXML(xml_node, slatTypes)
        self.ensure_3D()
        self.translation = self.getTranslation(True)
        self.ComputeBindableNbrGlues()
        self.compute_hash()
        self.outer_edges = None
     
    def __str__(self):
        return f'Slat of type {self.slatType.name}: blocks: {self.blocks}, domains: {self.domains}'

    def compute_hash(self):
        self.hash = str(self.translation) + self.slatType.name
        #self.hash = hash(self.translation) + hash(self.slatType.name)
        #print(f'hash: {self.hash}')

    def __hash__(self):
        print("ERROR: not implemented!")
        #return hash((self.slatType.name, self.get_reference_point_loc()))
        #return self.hash
        #=======================================================================
        # n = hash(self.slatType.name)
        # b = hash(tuple(sorted(list(self.blocks))))
        # d = hash(tuple(sorted([(l, tuple(sorted(d))) for l, d in self.domains.items()])))
        # return hash((n, b, d))
        #=======================================================================

    def __eq__(self, other):
        return self.hash == other.hash
        #=======================================================================
        # if self.slatType != other.slatType:
        #     return False
        # if self.domains.keys() != other.domains.keys():
        #     return False
        # for loc in self.domains:
        #     if set(self.domains[loc]) != set(other.domains[loc]):
        #         return False
        # if self.blocks != other.blocks:
        #     return False
        # return True
        #=======================================================================

    def ensure_3D(self):
        new_blocks = []
        for block in self.blocks:
            new_blocks.append(make_3D(block))
        self.blocks = new_blocks
        new_colors = dict()
        for color_coords in self.colors:
            new_colors[make_3D(color_coords)] = self.colors[color_coords]
        self.colors = new_colors
        new_domains = dict()
        for domain_coords in self.domains:
            new_domains[make_3D(domain_coords)] = self.domains[domain_coords]
        self.domains = new_domains

    def get_location_list(self):
        location_list = []
        if self.blocks:
            for block in self.blocks:
                location_list.append(block)
        return location_list

    def translate(self, vector):
        if len(vector) < 3:
            vector = (vector[0], vector[1], 0)
        new_blocks = []
        self.translation = (self.translation[0]+vector[0], self.translation[1]+vector[1], self.translation[2]+vector[2])
        for block in self.blocks:
            x,y,z = block[0] + vector[0], block[1] + vector[1], block[2] + vector[2]
            new_blocks.append((x,y,z))
        self.blocks = new_blocks

        translated_colors = dict()
        for color in self.colors:
            loc = (color[0] + vector[0], color[1] + vector[1], color[2] + vector[2])
            translated_colors[loc] = self.colors[color]
        self.colors = translated_colors

        translated_domains = dict()
        for domain in self.domains:
            loc = (domain[0] + vector[0], domain[1] + vector[1], domain[2] + vector[2])
            translated_domains[loc] = self.domains[domain]
        self.domains = translated_domains
        self.ComputeBindableNbrGlues()
        self.compute_hash()
        self.bounding_box = None

    def get_reference_point_loc(self):
        return (self.slatType.reference_point[0] + self.translation[0],
                self.slatType.reference_point[1] + self.translation[1],
                self.slatType.reference_point[2] + self.translation[2])
    
    def translate_reference_point_to(self, coords):
        diff = ( (coords[0] - self.slatType.reference_point[0]),
                 (coords[1] - self.slatType.reference_point[1]),
                 (coords[2] - self.slatType.reference_point[2]) )
        self.translate(diff)

    # Compute the "inverse" of this slat type's glues to make it easy to determine when binding can occur
    def ComputeBindableNbrGlues(self):
        self.bindable_nbr_glues = dict()
        for block in self.domains:
            domain_list = self.domains[block]
            for (label, direction, strength) in domain_list:
                domain_nbr_coords = (block[0]+DIR_OFFSETS[direction][0], block[1]+DIR_OFFSETS[direction][1], block[2]+DIR_OFFSETS[direction][2])
                domain_nbr_dir = DIR_OPPS[direction]
                domain_nbr_label = get_label_complement(label)
                domain_nbr_strength = strength
                self.bindable_nbr_glues[domain_nbr_coords] = (domain_nbr_label, domain_nbr_dir, domain_nbr_strength)

    def Draw(self, dc, zero_coords, side_len):
        for block in self.blocks:
            dc.SetBrush(wx.Brush(wx.TheColourDatabase.Find(self.color)))
            self.slatType.DrawBlock(dc, zero_coords, side_len, block, self.domains, self.colors)

    def Compute2DOutline(self, blocks=None):
        self.outer_edges = self.slatType.Get2DOutline(self.blocks)

    def Draw2DOutline(self, dc, zero_coords, side_len, blocks=None):
        global outline_pen
        if outline_pen is None:
            outline_pen = wx.Pen(wx.TheColourDatabase.Find("BLACK"), width=12)
        if self.outer_edges is None:
            self.Compute2DOutline()
        line_list = []
        for (start_pt, end_pt) in self.outer_edges:
            line_list.append((zero_coords[0]+(side_len*start_pt[0]),
                                zero_coords[1]-(side_len*start_pt[1]),
                                zero_coords[0]+(side_len*end_pt[0]),
                                zero_coords[1]-(side_len*end_pt[1])))
        saved_pen = dc.GetPen()
        dc.SetPen(outline_pen)
        dc.DrawLineList(line_list)
        dc.SetPen(saved_pen)

    def getDomainsAtLoc(self, coords):
        if coords in self.domains:
            return self.domains[coords]
        else:
            return None

    def getTranslation(self, b_compute = False):
        if not b_compute:
            return self.translation
        else:
            bError = False
            diff2 = None
            for bcount in range(len(self.blocks)):
                diff = (self.blocks[bcount][0]-self.slatType.blocks[bcount][0],
                        self.blocks[bcount][1]-self.slatType.blocks[bcount][1],
                        self.blocks[bcount][2]-self.slatType.blocks[bcount][2])
                if (not diff2 is None) and (not diff == diff2):
                    bError = True
                diff2 = diff
            if bError:
                print("ERROR: slat translation is not consistent")
            self.translation = diff
            return diff
        
    def getPerimeterLocations(self):
        perimeter_locs = []
        for po in self.slatType.perimeter_offsets:
            perimeter_locs.append((po[0]+self.translation[0], po[1]+self.translation[1], po[2]+self.translation[2]))
        return perimeter_locs
                                                                               
    def getBoundingBox(self):
        return self.slatType.getBoundingBox(self.blocks)

    def toXML(self, xml_parent=None):
        if xml_parent is None:
            xml_parent = ET.Element('Polyomino')
            elt_slat = xml_parent
        else:
            elt_slat = ET.SubElement(xml_parent, 'Polyomino')

        elt_slattype = ET.SubElement(elt_slat, 'PolyominoType')
        elt_slattype.text = self.slatType.name

        elt_translation = ET.SubElement(elt_slat, 'translation')
        elt_translation.text = str(self.getTranslation())

        return xml_parent

    def fromXML(self, xml_parent, slatTypes):
        if (not xml_parent.tag == 'Slat') and (not xml_parent.tag == 'Polyomino'):
            print("Invalid XML node for Slat")
        else:
            self.slatType = SlatType(None)
            blocks = []
            self.domains = dict()
            translation = (0,0,0)
            for elt_child in xml_parent:
                if (elt_child.tag == 'SlatType') or (elt_child.tag == 'PolyominoType'):
                    self.slatType = None
                    slatType_name = elt_child.text
                    if type(slatTypes) == list:
                        for pts in slatTypes:
                            if slatType_name in pts:
                                self.slatType = pts[slatType_name]
                    elif slatType_name in slatTypes:
                        self.slatType = slatTypes[slatType_name]
                    if self.slatType is None:
                        print(f"Error loading slat: Invalid slat type '{slatType_name}")
                        
                elif elt_child.tag == 'translation':
                    translation = eval(elt_child.text)
         
        if not self.slatType is None:
            self.color = self.slatType.color
            self.blocks = copy.deepcopy(self.slatType.blocks)
            self.colors = copy.deepcopy(self.slatType.colors)
            self.domains = copy.deepcopy(self.slatType.domains)
            self.translate(translation)
