import sys
import getopt
import os
import subprocess
import shutil
import glob
from SlatSystem import *


RODSIM_DIRECTORY = "/home/ubuntu/src/Wang/PolyominoSim/RodSim/build"
RODSIM_EXE = "rodsim_cli"

def DisplayHelp():
    print("\nThis script simulates input slat systems using rodsim and the parameters provided below, and possibly also")
    print("outputs the assemblies to PDF and/or SVG files.")
    print("\nUsage:")
    print("-h, --help\t\t\t\tDisplays this help message.")

    print("-x <xml file>,\t--xml=<xml file>\tXML file name of input file containing slat system.")
    print("\t\t\t\t\t Note that the xml_file parameter can also be a file name pattern to be matched.")

    print("-o <dir>,--out=<dir>\t\t\t[Optional] Directory for output files. If not specified,")
    print("\t\t\t\t\t the current directory will be used.")

    print("-p, --pdf\t\t\t\t[Optional] If this option is used, the slats will be output to a PDF file.")
    print("-s, --svg\t\t\t\t[Optional] If this option is used, the slats will be output to an SVG file.")
    print("-l <T/F>, --labels=<T/F>\t\t\t[Optional] Show lables in PDF/SVG, T for true, F for False (default T)")

    print("-m <mode>, --mode=<mode>\t\t[Optional] This sets the simulation mode, must be either 'abstract' or 'kinetic'.")
    print("\t\t\t\t\t If not specified, 'abstract' is used.")

    print("-n <num>, --numsteps=<num>\t\tSpecifies the number of steps for which to run each simulation.")

    print("-t=<num>, --threads=<num>\t\t[Optional] Specifies the number of parallel simulation threads to be run.")
    print("\t\t\t\t\t If not specified, defaults to 1.")

    print("-i <num>, --interval=<num>\t\t[Optional] Specifies how many simulation steps between outputting the assemblies.")
    print("\t\t\t\t\t If not specified, defaults to the value used for 'numsteps'.")

    print("-m <num>, --max=<num>\t\t\t[Optional] Specifies the maximum number of slat attachments per simulation.")
    exit()

def write_config(file_name, settings_dict):
    with open(file_name, 'w') as config_file:
        for setting in settings_dict:
            config_file.write(f'{setting}={settings_dict[setting]}\n')

if __name__ == "__main__":
    try:
        # starts at the second element of argv since the first one is the script name
        # extraparams are extra arguments passed after all option/keywords are assigned
        # opts is a list containing the pair "option"/"value"
        opts, extraparams = getopt.getopt(sys.argv[1:], "hx:pso:m:n:t:i:m:l:",
                            ["help", "xml=", "pdf", "svg", "out=", "mode=", "numsteps=", "threads=", "interval=", "max=", "labels="])
        #print 'Opts:',opts
        #print 'Extra parameters:',extraparams
    except:
        print("")
        print("Unrecognized option:" + str(sys.argv[1:]))
        DisplayHelp()

    if len(opts) == 0:
        DisplayHelp()

    sim_type = 'abstract'  # allowable values: 'abstract' and 'kinetic'
    xml_file_name = None
    b_pdf = False
    b_svg = False
    b_show_labels = True
    directory = '.'
    num_steps = 0
    ensemble_size = 1
    output_interval = None
    max_attachments = None

    for o,p in opts:
        if o in ['-h', '--help']:
            DisplayHelp()
        elif o in ['-x', '--xml']:
            xml_file_name = p
        elif o in ['-p', '--pdf']:
            b_pdf = True
        elif o in ['-s', '--svg']:
            b_svg = True
        elif o in ['-l', '--labels']:
            if (p == 'F') or (p == 'False'):
                b_show_labels = False
        elif o in ['-o', '--out']:
            directory = p
        elif o in ['-m', '--mode']:
            sim_type = p
        elif o in ['-n', '--numsteps']:
            num_steps = int(p)
        elif o in ['-t', '--threads']:
            ensemble_size = int(p)
        elif o in ['-i', '--interval']:
            output_interval = int(p)
        elif o in ['-m', '--max']:
            max_attachments = int(p)
        else:
            DisplayHelp()

    if not sim_type in ['abstract', 'kinetic']:
        DisplayHelp()

    if output_interval is None:
        output_interval = num_steps

    if xml_file_name is None:
        DisplayHelp()
    else:

        input_xml_file_list = glob.glob(xml_file_name)
        if len(input_xml_file_list) == 0:
            print(f"ERROR: No XML input files found matching name '{xml_file_name}'")
            exit(1)

        print(f'Preparing files for simulation(s):')
        # Create output directory (if necessary)
        #directory = os.path.abspath(directory)
        if not os.path.isdir(directory):
            make_dir = os.makedirs(directory)
            if not os.path.isdir(directory):
                print(f'ERROR: Failed to make output directory "{directory}". Exiting.')
                exit(1)
            else:
                print(f' Created output directory {directory}')

        # Copy simulator into output directory
        program = RODSIM_DIRECTORY
        if not program[-1] in ['/', '\\']:
            program += '/'
        program += RODSIM_EXE
        #program_copy = f'{directory}/{RODSIM_EXE}'
        #shutil.copy(program, program_copy)
        #print(f' Copied {program} to {program_copy}')
        program_copy = program

        print(f'Input XML files to be simulated: {input_xml_file_list}')
        for xml_file_name in input_xml_file_list:

            if not os.path.isfile(xml_file_name):
                print(f"ERROR: input xml file '{xml_file_name}' cannot be found.")
                exit(1)

            if '/' in xml_file_name:
                file_name_split = xml_file_name.split('/')
            elif '\\' in xml_file_name:
                file_name_split = xml_file_name.split('\\')
            else:
                file_name_split = [xml_file_name]
            output_base_name = file_name_split[-1][:-4]
            xml_file_name2 = file_name_split[-1]

            # Copy xml input file to output directory
            xml_file_copy_path = f'{directory}/{xml_file_name2}'
            shutil.copyfile(xml_file_name, xml_file_copy_path)
            print(f' Copied {xml_file_name} to {xml_file_copy_path}')

            # Create config file for simulator in output directory
            config_dict = dict()
            if num_steps is not None:
                config_dict['num_steps'] = num_steps
            if max_attachments is not None:
                config_dict['max_attachments'] = max_attachments
            if output_interval is not None:
                config_dict['output_interval'] = output_interval
            if ensemble_size is not None:
                config_dict['ensemble_size'] = ensemble_size
            if sim_type is not None:
                config_dict['sim_type'] = sim_type
            if xml_file_name is not None:
                config_dict['system'] = xml_file_name2
            if output_base_name is not None:
                config_dict['output_name'] = output_base_name
            #if directory is not None:
            #    config_dict['output_directory'] = directory

            if sim_type == 'abstract':
                # Reporting nondeterminism by default for abstract simulations
                config_dict['report_nondeterminism'] = True

            config_file_name = f'{directory}/{output_base_name}_config.txt'
            print(f" Writing config file to {config_file_name}")
            write_config(config_file_name, config_dict)

            # Get a list of file names in the output directory before running the simulator (to compare with afterward)
            xml_files_in_output_dir_before = glob.glob(f'{directory}/*xml')
            xml_file_time_dict = dict()
            for xml_file in xml_files_in_output_dir_before:
                xml_file_time_dict[xml_file] = os.path.getmtime(xml_file)

            # Run the simulator
            print(f'\nRunning simulator with config file {config_file_name}:')
            result = subprocess.run([program_copy, config_file_name], capture_output=True, text=True)
            print(f'\n Stdout: {result.stdout}')
            if len(result.stderr) > 0:
                print(f'ERRORS occurred!\n\n Stderr: {result.stderr}\n')

            # Check to see if we need to create PDF or SVG versions of assemblies
            if b_pdf or b_svg:
                # Get a list of new/updated xml files in the output directory
                xml_files_in_output_dir_after = glob.glob(f'{directory}/*xml')
                new_or_updated_xml_files = []
                for xml_file in xml_files_in_output_dir_after:
                    if not xml_file in xml_file_time_dict:
                        new_or_updated_xml_files.append(xml_file)
                    else:
                        timestamp = os.path.getmtime(xml_file)
                        if timestamp > xml_file_time_dict[xml_file]:
                            new_or_updated_xml_files.append(xml_file)
                #print(f'new_or_updated_xml_files: {new_or_updated_xml_files}')

                # Add input xml file for completeness
                new_or_updated_xml_files.append(xml_file_copy_path)

                if len(new_or_updated_xml_files) == 0:
                    print(' WARNING: No new assembly files created to draw to PDF or SVG')
                else:
                    if b_pdf and not b_svg:
                        print(' Drawing assemblies to PDF files')
                    elif b_svg and not b_pdf:
                        print(' Drawing assemblies to SVG files')
                    else:
                        print(' Drawing assemblies to PDF and SVG files')

                    for xml_file in new_or_updated_xml_files:
                        pdf_file_name = ''
                        if b_pdf:
                            pdf_file_name = f'{xml_file[:-4]}.pdf'
                        svg_file_name = ''
                        if b_svg:
                            svg_file_name = f'{xml_file[:-4]}.svg'

                        # Make images of output assemblies
                        slat_system = SlatSystem()
                        slat_system.ReadFromFile(xml_file)
                        display_mask = DISPLAY_ALL if b_show_labels else DISPLAY_NO_GLUES
                        slat_system.assembly.WriteToPDForSVG(pdfFileName=pdf_file_name, svgFileName=svg_file_name, displayMask=display_mask, sideLen=50)


        # Delete the copy of the simulator from the output directory when done
        if not program == program_copy:  # Make sure it's not the original copy first
            os.remove(program_copy)